/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __houdini_SurfaceAccessorHoudini_h__
#define __houdini_SurfaceAccessorHoudini_h__

#define	FE_SAH_ASSERT_BOUNDS	FALSE

namespace fe
{

// TODO rework select() methodology to be reentrant

/**************************************************************************//**
    @brief Houdini Surface Editing Implementation

	Note that attempts to access non-existant attributes isn't a failure.
	New attributes may be added when first written to by a set() method.

	@ingroup houdini
*//***************************************************************************/
class FE_DL_EXPORT SurfaceAccessorHoudini:
	public SurfaceAccessorBase
{
	public:
						SurfaceAccessorHoudini(void):
							m_pGdp(NULL),
							m_pGroup(NULL)									{}

virtual					~SurfaceAccessorHoudini(void)						{}

		BWORD			bind(SurfaceAccessibleI::Element a_element,
								SurfaceAccessibleI::Attribute a_attribute)
						{
							m_attribute=a_attribute;

							String name;
							switch(a_attribute)
							{
								case SurfaceAccessibleI::e_generic:
								case SurfaceAccessibleI::e_position:
									name="P";
									break;
								case SurfaceAccessibleI::e_normal:
									name="N";
									break;
								case SurfaceAccessibleI::e_vertices:
									FEASSERT(a_element==
											SurfaceAccessibleI::e_primitive);
									m_element=a_element;
									return TRUE;
							}
							return bindInternal(a_element,name);
						}
		BWORD			bind(SurfaceAccessibleI::Element a_element,
								const String& a_name)
						{
							m_attribute=SurfaceAccessibleI::e_generic;
							return bindInternal(a_element,a_name);
						}

						//* as SurfaceAccessorI
virtual	U32				count(void) const
						{
							FEASSERT(m_pGdp);

							if(m_attribute!=SurfaceAccessibleI::e_generic &&
									m_attribute!=
									SurfaceAccessibleI::e_vertices &&
									!isBound())
							{
								return 0;
							}

							switch(m_element)
							{
							case SurfaceAccessibleI::e_point:
								return m_pGdp->points().entries();

//							case SurfaceAccessibleI::e_vertex:
//								return 0;

							case SurfaceAccessibleI::e_primitive:
								return m_pGdp->primitives().entries();

//							case SurfaceAccessibleI::e_detail:
//								return 0;

							case SurfaceAccessibleI::e_pointGroup:
							case SurfaceAccessibleI::e_primitiveGroup:
								return m_pGroup? m_pGroup->entries(): 0;

							default:
								;
							}
							return 0;
						}
virtual	U32				subCount(U32 index) const
						{
							if(m_attribute!=SurfaceAccessibleI::e_generic &&
									m_attribute!=
									SurfaceAccessibleI::e_vertices &&
									!isBound())
							{
								return 0;
							}

							const_cast<SurfaceAccessorHoudini*>(this)->select(
									index);

							if(m_attribute==SurfaceAccessibleI::e_vertices)
							{
								FEASSERT(m_element==
										SurfaceAccessibleI::e_primitive);
								return m_pPrimitive->getVertexCount();
							}

							return 1;

							// TODO check what really needs sub-indexes
//							return m_handle.entries();
						}


virtual	void			set(U32 index,U32 subIndex,String a_string)
						{
#if	FE_SAH_ASSERT_BOUNDS
							FEASSERT(index<count());
							FEASSERT(subIndex<subCount(index));
#endif
							if(m_attribute!=SurfaceAccessibleI::e_vertices &&
									!isBound())
							{
								addString();
							}
							select(index);

							UT_String utString(a_string.c_str());
							m_handle.setString(utString,subIndex);
						}
virtual	String			string(U32 index,U32 subIndex=0)
						{
#if	FE_SAH_ASSERT_BOUNDS
							FEASSERT(index<count());
							FEASSERT(subIndex<subCount(index));
#endif
							select(index);

							if(m_attribute==SurfaceAccessibleI::e_vertices)
							{
								return String();
							}

							UT_String utString;
							m_handle.getString(utString,subIndex);
							return String(utString.buffer());
						}

virtual	void			set(U32 index,U32 subIndex,I32 a_integer)
						{
#if	FE_SAH_ASSERT_BOUNDS
							FEASSERT(index<count());
							FEASSERT(subIndex<subCount(index));
#endif
							if(m_attribute!=SurfaceAccessibleI::e_vertices &&
									!isBound())
							{
								addInteger();
							}
							select(index);

							m_handle.setI(a_integer,subIndex);
						}
virtual	I32				integer(U32 index,U32 subIndex=0)
						{
#if	FE_SAH_ASSERT_BOUNDS
							FEASSERT(index<count());
							FEASSERT(subIndex<subCount(index));
#endif
							select(index);

							if(m_element==SurfaceAccessibleI::e_pointGroup)
							{
								return m_pPoint->getNum();
							}

							if(m_element==
									SurfaceAccessibleI::e_primitiveGroup ||
									m_attribute==SurfaceAccessibleI::e_vertices)
							{
								FEASSERT(m_element==
										SurfaceAccessibleI::e_primitiveGroup
										|| m_element==
										SurfaceAccessibleI::e_primitive);
								return m_pPrimitive->getVertex(subIndex)
										.getBasePt()->getNum();
							}

							return m_handle.getI(subIndex);
						}

virtual	void			set(U32 index,U32 subIndex,Real a_real)
						{
#if	FE_SAH_ASSERT_BOUNDS
							FEASSERT(index<count());
							FEASSERT(subIndex<subCount(index));
#endif
							if(m_attribute!=SurfaceAccessibleI::e_vertices &&
									!isBound())
							{
								addReal();
							}
							select(index);

							m_handle.setF(a_real,subIndex);
						}
virtual	Real			real(U32 index,U32 subIndex=0)
						{
#if	FE_SAH_ASSERT_BOUNDS
							FEASSERT(index<count());
							FEASSERT(subIndex<subCount(index));
#endif
							select(index);

							if(m_attribute==SurfaceAccessibleI::e_vertices)
							{
								return 0.0;
							}

							return m_handle.getF(subIndex);
						}

virtual	void			set(U32 index,U32 subIndex,
								const SpatialVector& a_vector)
						{
#if	FE_SAH_ASSERT_BOUNDS
							FEASSERT(index<count());
							FEASSERT(subIndex<subCount(index));
#endif
							if(m_element!=SurfaceAccessibleI::e_pointGroup &&
									m_element!=
									SurfaceAccessibleI::e_primitiveGroup &&
									m_attribute!=
									SurfaceAccessibleI::e_vertices &&
									!isBound())
							{
								addSpatialVector();
							}
							select(index);

							if(m_element==SurfaceAccessibleI::e_pointGroup)
							{
								UT_Vector4 utVector4(a_vector.temp());
								m_pPointChangeable->setPos(utVector4);
								return;
							}

							if(m_element==
									SurfaceAccessibleI::e_primitiveGroup ||
									m_attribute==SurfaceAccessibleI::e_vertices)
							{
								FEASSERT(m_element==
										SurfaceAccessibleI::e_primitiveGroup
										|| m_element==
										SurfaceAccessibleI::e_primitive);
								UT_Vector4 utVector4(a_vector.temp());
								m_pPrimitive->getVertex(subIndex).getPt()
										->setPos(utVector4);
								return;
							}

							UT_Vector4 utVector4(a_vector.temp());
							m_handle.setV4(utVector4);
						}
virtual	SpatialVector	spatialVector(U32 index,U32 subIndex=0)
						{
#if	FE_SAH_ASSERT_BOUNDS
							FEASSERT(index<count());
							FEASSERT(subIndex<subCount(index));
#endif
							select(index);

							if(m_element==SurfaceAccessibleI::e_pointGroup)
							{
								UT_Vector4 utVector4=m_pPoint->getPos();
								SpatialVector vector(utVector4.data());
								return vector;
							}

							if(m_element==
									SurfaceAccessibleI::e_primitiveGroup ||
									m_attribute==SurfaceAccessibleI::e_vertices)
							{
								FEASSERT(m_element==
										SurfaceAccessibleI::e_primitiveGroup
										|| m_element==
										SurfaceAccessibleI::e_primitive);
								UT_Vector4 utVector4=m_pPrimitive
										->getVertex(subIndex).getPos();
								SpatialVector vector(utVector4.data());
								return vector;
							}

							UT_Vector4 utVector4=m_handle.getV4(subIndex);
							SpatialVector vector(utVector4.data());
							return vector;
						}

						//* Houdini specific
		void			setGdp(const GU_Detail* a_pGdp)
						{
							m_pGdp=a_pGdp;
							m_pGdpChangeable=NULL;
						}
		void			setGdpChangeable(GU_Detail* a_pGdp)
						{
							m_pGdp=a_pGdp;
							m_pGdpChangeable=a_pGdp;
						}
const	GU_Detail*		gdp(void)				{ return m_pGdp; }
		GU_Detail*		gdpChangeable(void)		{ return m_pGdpChangeable; }

	private:
		void			addString(void);
		void			addInteger(void);
		void			addReal(void);
		void			addSpatialVector(void);

		BWORD			isBound(void) const
						{	return (m_handle.isAttributeValid() || m_pGroup); }

		BWORD			bindInternal(SurfaceAccessibleI::Element a_element,
								const String& a_name);
		void			select(U32 index);

		SurfaceAccessibleI::Element		m_element;
		SurfaceAccessibleI::Attribute	m_attribute;
		String							m_name;	//* in case we need to add it

const	GU_Detail*						m_pGdp;
		GU_Detail*						m_pGdpChangeable;

const	GEO_Point*						m_pPoint;
		GEO_Point*						m_pPointChangeable;
const	GEO_Primitive*					m_pPrimitive;
		GEO_AttributeHandle				m_handle;

#ifdef FE_HOUDINI_USE_GA
		GA_Group*						m_pGroup;
#else
		GB_Group*						m_pGroup;
#endif
		U32								m_groupIndex;
};

} // namespace

#endif /* __houdini_SurfaceAccessorHoudini_h__ */
