/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_GWO_DEBUG	FALSE

#include "operator/operator.pmh"

using namespace fe;

void GridWrapOp::initialize(void)
{
	catalog<String>("icon")="FE_alpha";

	//* don't show thread attributes until we start using them
	catalog<bool>("Threads","hidden")=true;
	catalog<bool>("StayAlive","hidden")=true;

	catalog<I32>("Grid Width")=10;
	catalog<I32>("Grid Width","high")=100;
	catalog<I32>("Grid Width","max")=1000;
	catalog<bool>("Grid Width","join")=true;

	catalog<I32>("Grid Height")=10;
	catalog<I32>("Grid Height","high")=100;
	catalog<I32>("Grid Height","max")=1000;

	catalog<I32>("Convolutions")=1;
	catalog<I32>("Convolutions","high")=10;
	catalog<I32>("Convolutions","max")=100;

	catalog<String>("Coordinate Attribute")="uv";

	catalog<bool>("Debug")=false;

	catalog< sp<Component> >("Input Surface");

	catalog< sp<Component> >("Driver Surface");

	catalog< sp<Component> >("Deformed Surface");

	catalog< sp<Component> >("Output Surface");
	catalog<String>("Output Surface","IO")="output";
	catalog<String>("Output Surface","copy")="Input Surface";
	catalog<bool>("Output Surface","recycle")=true;
}

void GridWrapOp::handle(Record& a_rSignal)
{
#if FE_GWO_DEBUG
	feLog("GridWrapOp::handle\n");
#endif

	sp<DrawI> spDrawDebug;
	if(catalog<BWORD>("Debug") && !accessGuide(spDrawDebug,a_rSignal)) return;

	sp<SurfaceAccessibleI> spInputAccessible;
	if(!access(spInputAccessible,"Input Surface")) return;

	sp<SurfaceI> spDriver;
	if(!access(spDriver,"Driver Surface")) return;

	spDriver->setRefinement(0);

	sp<SurfaceI> spDeformed;
	if(!access(spDeformed,"Deformed Surface")) return;

	spDeformed->setRefinement(0);

	sp<SurfaceAccessorI> spInputPoint;
	if(!access(spInputPoint,spInputAccessible,e_point,e_position)) return;

	sp<SurfaceAccessorI> spOutputPoint;
	if(!accessOutput(spOutputPoint,a_rSignal,e_point,e_position)) return;

	const U32 count=spInputPoint->count();
	if(!count)
	{
		feLog("GridWrapOp input point count is zero\n");
		catalog<String>("error")="input points required";
		return;
	}

	const String coordinateAttribute=catalog<String>("Coordinate Attribute");

	sp<SurfaceAccessorI> spInputUV;
	if(!access(spInputUV,spInputAccessible,e_point,coordinateAttribute)) return;

	const I32 gridWidth=catalog<I32>("Grid Width");
	const I32 gridHeight=catalog<I32>("Grid Height");
	const I32 convolutions=catalog<I32>("Convolutions");

	const U32 gridCount=gridWidth*gridHeight;
	GridNode grid[2][gridCount];
	U32 writeIndex=0;
	GridNode* writeGrid=grid[writeIndex];
	GridNode* readGrid=grid[!writeIndex];

	for(U32 gridIndex=0;gridIndex<gridCount;gridIndex++)
	{
		setIdentity(writeGrid[gridIndex].m_transform);
		setIdentity(readGrid[gridIndex].m_transform);
	}

	for(U32 index=0;index<count;index++)
	{
		const SpatialVector inputPoint=spInputPoint->spatialVector(index);
		const SpatialVector inputUV=spInputUV->spatialVector(index);

		sp<SurfaceI::ImpactI> spImpact=spDriver->nearestPoint(inputPoint);
		if(!spImpact.isValid())
		{
			continue;
		}

		I32 face=spImpact->face();
		Barycenter<Real> barycenter=spImpact->barycenter();

		const SpatialTransform xformDriver=
				spDriver->sample(face,barycenter);
		const SpatialTransform xformDeformed=
				spDeformed->sample(face,barycenter);
		SpatialTransform invDriver;
		invert(invDriver,xformDriver);
		const SpatialTransform conversion=invDriver*xformDeformed;

		I32 gridU=gridWidth*inputUV[0]-1e-3;
		I32 gridV=gridHeight*inputUV[1]-1e-3;

		if(gridU<0 || gridU>=gridWidth || gridV<0 || gridV>=gridHeight)
		{
			continue;
		}

		//* TODO cumulative
		writeGrid[gridWidth*gridV+gridU].m_transform=conversion;
		writeGrid[gridWidth*gridV+gridU].m_weight+=1.0;

//		feLog("GridWrapOp::handle scan %d,%d\n%s\n",
//				gridU,gridV,
//				c_print(writeGrid[gridWidth*gridV+gridU].m_transform));

		if(spDrawDebug.isValid())
		{
			const Color yellow(1.0,1.0,0.0,1.0);
			SpatialVector line[2];

			line[0]=xformDriver.translation();
			line[1]=xformDeformed.translation();

			spDrawDebug->drawLines(line,NULL,2,DrawI::e_strip,false,&yellow);
		}
	}

	for(U32 convolution=0;convolution<convolutions;convolution++)
	{
		writeIndex=!writeIndex;
		writeGrid=grid[writeIndex];
		readGrid=grid[!writeIndex];

		for(I32 gridV=0;gridV<gridHeight;gridV++)
		{
			const I32 gridVV=gridV*gridWidth;
			for(I32 gridU=0;gridU<gridWidth;gridU++)
			{
				GridNode& rGridWrite=writeGrid[gridVV+gridU];

				SpatialVector sumTangent;
				SpatialVector sumNormal;
				SpatialVector sumLocation;
				Real sumWeight=0.0;

				set(sumTangent);
				set(sumNormal);
				set(sumLocation);

#if FE_GWO_DEBUG
				feLog("GridWrapOp::handle convolution %d %d/%d %d/%d\n",
						convolution,gridU,gridWidth,gridV,gridHeight);
#endif

				const I32 maxV=gridV<gridHeight-1? 2: 1;
				for(I32 kernelV=(gridV? -1: 0);kernelV<maxV;kernelV++)
				{
					const I32 kernelVV=gridVV+kernelV*gridWidth+gridU;

					const I32 maxU=gridU<gridWidth-1? 2: 1;
					for(I32 kernelU=(gridU? -1: 0);kernelU<maxU;kernelU++)
					{
						GridNode& rGridRead=readGrid[kernelVV+kernelU];
						SpatialTransform& rTransform=rGridRead.m_transform;

						if(rGridRead.m_weight<0.0)
						{
#if FE_GWO_DEBUG
							feLog("GridWrapOp::handle"
									" convolution %d,%d %d,%d SKIP\n",
									gridU,gridV,kernelU,kernelV);
#endif
							continue;
						}

#if FE_GWO_DEBUG
						feLog("GridWrapOp::handle"
								" convolution %d,%d %d,%d\n%s\n",
								gridU,gridV,kernelU,kernelV,
								c_print(rTransform));
#endif

						sumTangent+=rTransform.direction();
						sumNormal+=rTransform.left();
						sumLocation+=rTransform.translation();
						sumWeight+=1.0;

#if FE_GWO_DEBUG
						feLog("GridWrapOp::handle"
								" weight %.6G\n"
								" sumTangent %s\n"
								" sumNormal %s\n"
								" sumLocation %s\n",
								sumWeight,
								c_print(sumTangent),
								c_print(sumNormal),
								c_print(sumLocation),
								c_print(sumTangent));
#endif
					}
				}

				if(sumWeight>1.0)
				{
					const Real invWeight=1.0/sumWeight;
					const BWORD success=makeFrameNormalY(
							rGridWrite.m_transform,sumLocation*invWeight,
							sumTangent,unitSafe(sumNormal));
					if(!success)
					{
						feLog("GridWrapOp::handle"
								" convolution makeFrameNormalY failed %d,%d\n",
								gridU,gridV);
					}
					rGridWrite.m_weight=Real(sumWeight);
				}
				else
				{
					setIdentity(rGridWrite.m_transform);
					rGridWrite.m_weight=Real(0);
				}

#if FE_GWO_DEBUG
				feLog("GridWrapOp::handle result\n%s\n",
						c_print(rGridWrite.m_transform));
#endif
			}
		}
	}

	writeIndex=!writeIndex;
	writeGrid=grid[writeIndex];
	readGrid=grid[!writeIndex];

	for(U32 index=0;index<count;index++)
	{
		const SpatialVector inputPoint=spInputPoint->spatialVector(index);
		const SpatialVector inputUV=spInputUV->spatialVector(index);

		I32 gridU=gridWidth*inputUV[0]-1e-3;
		I32 gridV=gridHeight*inputUV[1]-1e-3;

		if(gridU<0 || gridU>=gridWidth || gridV<0 || gridV>=gridHeight)
		{
			continue;
		}

		//* TODO blend
		const SpatialTransform& rConversion=
				readGrid[gridWidth*gridV+gridU].m_transform;

		SpatialVector transformed;
		transformVector(rConversion,inputPoint,transformed);

		spOutputPoint->set(index,transformed);
	}

#if FE_GWO_DEBUG
	feLog("GridWrapOp::handle done\n");
#endif
}
